
#include "dualhierarchy.h"
#include <fstream>

dualhierarchy::dualhierarchy(pattern ** xmlQueries, int xmlCounter)
{
	xmlArray = xmlQueries;
	xmlSize = xmlCounter;

	xmlStarredArray = new pattern*[xmlSize];
	for (int i = 0; i < xmlSize; i++){
		xmlStarredArray[i] = new pattern(xmlArray[i]->getXPath(), xmlArray[i]->getIndex(), xmlArray[i]->getQueryNumber(), xmlArray[i]->getXPathType());
		xmlStarredArray[i]->addSlashSlashStar();
	}

	subPatterns = new patternlist*[xmlSize];
	equivPatterns = new patternlist*[xmlSize];
	shouldPrint = new int[xmlSize];
	edgeShouldPrint = new bool*[xmlSize];
	
	for (int q = 0; q < xmlSize; q++){
		subPatterns[q] = NULL;
		equivPatterns[q] = NULL;
		shouldPrint[q] = -1;
		edgeShouldPrint[q] = new bool[xmlSize];
		for (int w = 0; w < xmlSize; w++){
			edgeShouldPrint[q][w] = true;
		}

	}
	treeRoots = NULL;
};

dualhierarchy::~dualhierarchy()
{
	for (int q = 0; q < xmlSize; q++){
		if(subPatterns[q] != NULL){
			delete subPatterns[q];
		}
		if(equivPatterns[q] != NULL){
			delete subPatterns[q];
		}
	}

	delete treeRoots;
	delete [] subPatterns;
	delete [] equivPatterns;

};


void dualhierarchy::eliminateRedundantEdges(){

	bool firstLevelIsContain;
	bool secondLevelIsContain;
	bool altLevelIsContain;
	for(int i = 0; i < xmlSize; i++){
		if(subPatterns[i] != NULL){
			if(shouldPrint[i] == -1){
				patternlist* itr1 = subPatterns[i];
				pattern* sub1 = itr1->getCurrent();
				while (itr1 != NULL){
					pattern* sub1 = itr1->getCurrent();
					patternlist* itr2 = subPatterns[sub1->getIndex()];
					firstLevelIsContain = itr1->getContainBool();
					if(shouldPrint[sub1->getIndex()] == -1){
						while (itr2 != NULL){
							pattern* sub2 = itr2->getCurrent();
							secondLevelIsContain = itr2->getContainBool();
							if(shouldPrint[sub2->getIndex()] == -1){
								if(itr1 != itr2){
									patternlist* itr3 = subPatterns[i];
									while(itr3 != NULL){
										pattern* sub3 = itr3->getCurrent();
										altLevelIsContain = itr3->getContainBool();
										if(sub3 == sub2){
											if((firstLevelIsContain && !secondLevelIsContain) || 
											(!firstLevelIsContain && secondLevelIsContain)){
												if(!altLevelIsContain){
													edgeShouldPrint[i][sub3->getIndex()] = false;
												}
											}else if(firstLevelIsContain && secondLevelIsContain){
												if(altLevelIsContain){
													edgeShouldPrint[i][sub3->getIndex()] = false;
												}
											}else if(!firstLevelIsContain && !secondLevelIsContain){
												if(!altLevelIsContain){
													edgeShouldPrint[i][sub3->getIndex()] = false;
												}
											}
										}
										itr3 = itr3->getNext();
									}
								}
							}
							itr2 = itr2->getNext();
						}
					}
					itr1 = itr1->getNext();
				}
			}
		}
	}
};



void dualhierarchy::sameHierarchyLevelContainmentFinder(){

	for(int i = 0; i < xmlSize; i++){
		if(subPatterns[i] != NULL){
			if(shouldPrint[i] == -1){
				patternlist* itr1 = subPatterns[i];
				pattern* sub1 = itr1->getCurrent();
				while (itr1 != NULL){
					patternlist* itr2 = subPatterns[i];
					pattern* sub2 = itr2->getCurrent();
					pattern* sub1 = itr1->getCurrent();
					if(!itr1->getContainBool() && shouldPrint[itr1->getCurrent()->getIndex()] == -1){
						while (itr2 != NULL){
							pattern* sub2 = itr2->getCurrent();
							if(!itr2->getContainBool() && shouldPrint[itr2->getCurrent()->getIndex()] == -1){
								if(itr1 != itr2){
									if (sub1->DynamicProgrammingAlgorithm(sub2, true)){
										if(subPatterns[sub1->getIndex()] == NULL){
											subPatterns[sub1->getIndex()] = new patternlist(sub2, true);
										}else{
											subPatterns[sub1->getIndex()]->addItem(sub2,true);
										}	
									}else{
										checkContainInSubtree(sub1, subPatterns[sub2->getIndex()]);
									}
								}
							}
							itr2 = itr2->getNext();
						}
					}
					itr1 = itr1->getNext();
				}
			}
		}
	}
};


void dualhierarchy::checkContainInSubtree(pattern* subTop, patternlist* subBottomList){
	pattern* subBottom;
	while (subBottomList != NULL){
		subBottom = subBottomList->getCurrent();
		if(!subBottomList->getContainBool() && shouldPrint[subBottom->getIndex()] == -1){
			if(subTop != subBottom){
				if (subTop->DynamicProgrammingAlgorithm(subBottom, true)){
					if(subPatterns[subTop->getIndex()] == NULL){
						subPatterns[subTop->getIndex()] = new patternlist(subBottom, true);
					}else{
						subPatterns[subTop->getIndex()]->addItem(subBottom,true);
					}	
				}else{
					checkContainInSubtree(subTop, subPatterns[subBottom->getIndex()]);
				}
			}
		}
		subBottomList = subBottomList->getNext();
	}
}

void dualhierarchy::addPattern(int x)
{
	addPatternHelper(x, treeRoots, false);
};

void dualhierarchy::addPatternHelper(int x, patternlist* &rootPassedIn, bool forcedContain)
{
	patternlist* rootCounter = rootPassedIn;
	patternlist* prevCounter = NULL;
	bool addedYet = false;
	bool quit = false;
	while(rootCounter != NULL && !quit){
		bool newStarredIsBigger = xmlStarredArray[x]->DynamicProgrammingAlgorithm(rootCounter->getCurrent(), altResults); // false
		bool newStarredIsSmaller = xmlStarredArray[rootCounter->getCurrent()->getIndex()]->DynamicProgrammingAlgorithm(xmlArray[x], altResults); // false
		bool newIsBigger = xmlArray[x]->DynamicProgrammingAlgorithm(xmlArray[rootCounter->getCurrent()->getIndex()], true);
		bool newIsSmaller = xmlArray[rootCounter->getCurrent()->getIndex()]->DynamicProgrammingAlgorithm(xmlArray[x], true);

		if(newIsSmaller && newIsBigger){
			// equivalent!
			if(equivPatterns[rootCounter->getCurrent()->getIndex()] == NULL){
				equivPatterns[rootCounter->getCurrent()->getIndex()] = new patternlist(xmlArray[x], true);
			}else{
				equivPatterns[rootCounter->getCurrent()->getIndex()]->addItem(xmlArray[x], true);
			}
			shouldPrint[x] = rootCounter->getCurrent()->getIndex();
			addedYet = true;
		}else if(newStarredIsBigger && !newStarredIsSmaller){

			subPatterns[x] = new patternlist(rootCounter->getCurrent(), newIsBigger);
			rootCounter->setCurrent(xmlArray[x], forcedContain); // rootCounter->getContainBool()
			prevCounter = rootCounter;
			rootCounter = rootCounter->getNext();
			while (rootCounter != NULL){
				newStarredIsBigger = xmlStarredArray[x]->DynamicProgrammingAlgorithm(rootCounter->getCurrent(), altResults); // false
				if(newStarredIsBigger){
					newIsBigger = xmlArray[x]->DynamicProgrammingAlgorithm(xmlArray[rootCounter->getCurrent()->getIndex()], true);
					subPatterns[x]->addItem(rootCounter->getCurrent(), newIsBigger);
					// prevCounter should never be null
					prevCounter->removeNextFromList();
					rootCounter = prevCounter->getNext();
				}else{
					prevCounter = rootCounter;
					rootCounter = rootCounter->getNext();
				}
			}

			addedYet = true;

		}else if(newStarredIsSmaller && !newStarredIsBigger){
			addPatternHelper(x,subPatterns[rootCounter->getCurrent()->getIndex()], newIsSmaller);
			addedYet = true;
		}else if(newStarredIsSmaller && newStarredIsBigger){
			// equivalent!
			if(equivPatterns[rootCounter->getCurrent()->getIndex()] == NULL){
				equivPatterns[rootCounter->getCurrent()->getIndex()] = new patternlist(xmlArray[x], true);
			}else{
				equivPatterns[rootCounter->getCurrent()->getIndex()]->addItem(xmlArray[x], true);
			}
			shouldPrint[x] = rootCounter->getCurrent()->getIndex();
			addedYet = true;
		}else if(newIsBigger && !altResults){
			subPatterns[x] = new patternlist(rootCounter->getCurrent(), newIsBigger);
			rootCounter->setCurrent(xmlArray[x], rootCounter->getContainBool());
			prevCounter = rootCounter;
			rootCounter = rootCounter->getNext();
			while (rootCounter != NULL){
				newStarredIsBigger = xmlStarredArray[x]->DynamicProgrammingAlgorithm(rootCounter->getCurrent(), altResults); // false
				if(newStarredIsBigger){
					newIsBigger = xmlArray[x]->DynamicProgrammingAlgorithm(xmlArray[rootCounter->getCurrent()->getIndex()], true);
					subPatterns[x]->addItem(rootCounter->getCurrent(), newIsBigger);
					// prevCounter should never be null
					prevCounter->removeNextFromList();
					rootCounter = prevCounter->getNext();
				}else{
					prevCounter = rootCounter;
					rootCounter = rootCounter->getNext();
				}
			}

			addedYet = true;
		}else if(newIsSmaller && !altResults){
			addPatternHelper(x,subPatterns[rootCounter->getCurrent()->getIndex()], true);
			addedYet = true;
		}else{
			// still have to call single-check function
		//	singleCheck(x, subPatterns[rootCounter->getCurrent()->getIndex()]);
		}
		prevCounter = rootCounter;
		if(rootCounter != NULL){
			rootCounter = rootCounter->getNext();
		}
	}
	if(!addedYet){
		if(prevCounter != NULL){
			prevCounter->addItem(xmlArray[x], forcedContain);
		}else{
			rootPassedIn = new patternlist(xmlArray[x], forcedContain);
		}
	}
};
/*
void dualhierarchy::singleCheck(int x, patternlist* rootCounter)
{
//	patternlist* rootCounter = rootPassedIn;

	while (rootCounter != NULL){
		if(xmlStarredArray[x]->DynamicProgrammingAlgorithm(rootCounter->getCurrent(), false) ){
			if(subPatterns[x] == NULL){
				subPatterns[x] = new patternlist(rootCounter->getCurrent());
			}else{
				subPatterns[x]->addItem(rootCounter->getCurrent());
			}
		}else{
			singleCheck(x, subPatterns[rootCounter->getCurrent()->getIndex()]);
		}
		rootCounter = rootCounter->getNext();
	}
};
*/

void dualhierarchy::printToDot(string filename, bool xPathsAsNumbers, bool showSource, bool collapseNodes, bool onlyXPaths, int width, int height)
{

	ofstream dotFile;
	dotFile.open(filename.c_str());

//	for (int i = 0; i < xmlSize; i++){
//		dotFile << "//:: " << i << " ::: " << xmlArray[i]->getXPath() << endl;
//	}

	dotFile << "digraph XPath_Query_Relationships {" << endl << endl;

//	dotFile << "graph [rankdir = \"LR\"];" << endl;

//	dotFile << "size=\"" << width << ", " << height << "\"" << endl;
	int i;
	if(onlyXPaths){
		for (i = 0; i < xmlSize; i++){
			if(shouldPrint[i] < 0){
				dotFile << i << " [shape=ellipse," << "label=\"" << xmlArray[i]->getPrintableXPath() << "\"];" << endl;
			}
		}
	}else{
		for (i = 0; i < xmlSize; i++){
			if(shouldPrint[i] < 0){
				int fieldLabel = 0;
				dotFile << i << " [shape=record," << "label=\"{<f" << fieldLabel << "1>"; //{{<f
				if(xPathsAsNumbers){
					dotFile << "XPaths: " << i;
				}else{
					dotFile << xmlArray[i]->getPrintableXPath();
				}
				if(collapseNodes){
					bool* printedQueries = new bool[xmlSize];
					for(int c = 0; c< xmlSize; c++){
						printedQueries[c] = false;
					}

					if(xPathsAsNumbers){
						patternlist* equivs = equivPatterns[i];
						while (equivs != NULL) {
							dotFile << ", " << equivs->getCurrent()->getIndex();
							equivs = equivs->getNext();
						}
					}
					if(xmlArray[i]->getQueryNumber() >= 0){
						int queryNum = xmlArray[i]->getQueryNumber();
						queryNum++;
						//if(queryNum > 6){
						//	queryNum++;
						//}
						
						dotFile << "|<f" << fieldLabel << "2>XQueries: " << queryNum; //}|<f"
						printedQueries[queryNum] = true;
					}
					patternlist* equivs = equivPatterns[i];

					while (equivs != NULL) {
						if(xmlArray[equivs->getCurrent()->getIndex()]->getQueryNumber() >= 0){
							if(xmlArray[i]->getQueryNumber() >= 0){
								int queryNum = xmlArray[equivs->getCurrent()->getIndex()]->getQueryNumber();
								queryNum++;
								//if(queryNum > 6){
								//	queryNum++;
								//}

								if(!printedQueries[queryNum]){
									dotFile << ", " << queryNum;
									printedQueries[queryNum] = true;
								}
							}
						}
						equivs = equivs->getNext();
					}

				}else{
					if(xmlArray[i]->getQueryNumber() >= 0){
						int queryNum = xmlArray[i]->getQueryNumber();

						dotFile << "|<f" << fieldLabel << "2>XQuery ID: " << queryNum;
						if(showSource){
							dotFile << "|<f" << fieldLabel << "3>XQuery Part: " << xmlArray[i]->getXPathType();
						}
						//dotFile << "}";
					}else{
						//dotFile << "}";
					}

					patternlist* equivs = equivPatterns[i];
					while (equivs != NULL) {
						fieldLabel++;
						dotFile << "|{<f" << fieldLabel << "1>";
						if(xPathsAsNumbers){
							dotFile << "XPath ID: " << equivs->getCurrent()->getIndex();
						}else{
							dotFile << xmlArray[equivs->getCurrent()->getIndex()]->getPrintableXPath();
						}
						if(xmlArray[i]->getQueryNumber() >= 0){
							dotFile << "|<f" << fieldLabel << "2>XQuery ID: " << xmlArray[equivs->getCurrent()->getIndex()]->getQueryNumber();
							if(showSource){
								dotFile << "|<f" << fieldLabel << "3>XQuery Part: " << xmlArray[equivs->getCurrent()->getIndex()]->getXPathType();
							}
							//dotFile << "}";
						}else{
							//dotFile << "}";
						}
						equivs = equivs->getNext();
					}
				}

				dotFile << "}\"];" << endl; // "}\"];"
			}
		}
	}

	dotFile << endl;

	patternlist * rootItr = treeRoots;
//	while(rootItr !=NULL){}

	bool **addedEdge = new bool*[xmlSize];
	for (int n = 0; n < xmlSize; n++){
		addedEdge[n] = new bool[xmlSize];
		for (int m=0; m < xmlSize; m++){
			addedEdge[n][m] = false;
		}
	}

	for (int j = 0; j < xmlSize; j++){
		
		rootItr = subPatterns[j];
		int first;
		int last;
		while(rootItr != NULL){
			if(shouldPrint[j] < 0){
				first = j;
			}else{
				first = shouldPrint[j];
			}
			if(shouldPrint[rootItr->getCurrent()->getIndex()] < 0){
				last = rootItr->getCurrent()->getIndex();
			}else{
				last = shouldPrint[rootItr->getCurrent()->getIndex()];
			}
			
			if(!addedEdge[first][last]  && edgeShouldPrint[first][last]){
				dotFile << first << " -> " << last; 
				if(!rootItr->getContainBool()){
					dotFile << " [style=\"bold\"]";
				}else{
					dotFile << " [style=\"dashed\"]";
				}
				dotFile << ";" << endl;
				addedEdge[first][last] = true;
			}
			rootItr = rootItr->getNext();
		}
	}

	for (int b = 0; b < xmlSize; b++){
		delete addedEdge[b];
	}
	delete addedEdge;

/*
			rootItr = equivPatterns[j];
			while(rootItr != NULL){
				if(shouldPrint[j] < 0){
					first = j;
				}else{
					first = shouldPrint[j];
				}
				if(shouldPrint[rootItr->getCurrent()->getIndex() < 0){
					last = rootItr->getCurrent()->getIndex()
				}else{
					last = shouldPrint[rootItr->getCurrent()->getIndex()];
				}
				dotFile << first << " -> " << last << ";" << endl;
				dotFile << last << " -> " << first << ";" << endl;
				rootItr = rootItr->getNext();
			}

*/

	dotFile << endl << "}";
	dotFile.close();

};



void dualhierarchy::printPathsToDot(string filename)
{

	ofstream dotFile;
	dotFile.open(filename.c_str());

	for (int i = 0; i < xmlSize; i++){
		dotFile << "//:: " << i << " ::: " << xmlArray[i]->getXPath() << endl;
	}

	dotFile << "digraph XPath_Query_Relationships {" << endl << endl;

	dotFile << "size=\"" << (xmlSize / 6) << ", " << (xmlSize / 6) << "\"" << endl;

	dotFile << "1000 [shape=record," <<"label=\"{ ID " ;
	int x;
	for (x = 0; x < xmlSize; x++){
		dotFile << " | " << x ;
	}
	dotFile << "} | { " << "XPath" ;
	for (x = 0; x < xmlSize; x++){
		dotFile << " | " << xmlArray[x]->getPrintableXPath() ;
	}
	dotFile << "} | { " << "XQuery" ;
	for (x = 0; x < xmlSize; x++){
		dotFile << " | " << xmlArray[x]->getQueryNumber() ;
	}
	dotFile << "} | { " << "Query Part" ;
	for (x = 0; x < xmlSize; x++){
		dotFile << " | " << xmlArray[x]->getXPathType() ;
	}
	dotFile << "}\"];" << endl;
	dotFile << "}";
	dotFile.close();

};
