
#include "hierarchy.h"
#include <fstream>


hierarchy::hierarchy(pattern ** xmlQueries, int xmlCounter)
{
	xmlArray = xmlQueries;
	xmlSize = xmlCounter;
	subPatterns = new patternlist*[xmlSize];
	equivPatterns = new patternlist*[xmlSize];
	shouldPrint = new int[xmlSize];
	
	for (int q = 0; q < xmlSize; q++){
		subPatterns[q] = NULL;
		equivPatterns[q] = NULL;
		shouldPrint[q] = -1;
	}
	treeRoots = NULL;
};

hierarchy::~hierarchy()
{
	for (int q = 0; q < xmlSize; q++){
		if(subPatterns[q] != NULL){
			delete subPatterns[q];
		}
		if(equivPatterns[q] != NULL){
			delete subPatterns[q];
		}
	}

	delete treeRoots;
	delete [] subPatterns;
	delete [] equivPatterns;

};

void hierarchy::addPattern(int x)
{
	addPatternHelper(x, treeRoots);
};

void hierarchy::addPatternHelper(int x, patternlist* &rootPassedIn)
{
	patternlist* rootCounter = rootPassedIn;
	patternlist* prevCounter = NULL;
	bool addedYet = false;
	bool quit = false;
	while(rootCounter != NULL && !quit){
		bool newIsBigger = xmlArray[x]->DynamicProgrammingAlgorithm(rootCounter->getCurrent(), true);
		bool newIsSmaller = rootCounter->getCurrent()->DynamicProgrammingAlgorithm(xmlArray[x], true);
		if(newIsBigger && !newIsSmaller){
			subPatterns[x] = new patternlist(rootCounter->getCurrent());
			rootCounter->setCurrent(xmlArray[x]);
			prevCounter = rootCounter;
			rootCounter = rootCounter->getNext();
			while (rootCounter != NULL){
				if(xmlArray[x]->DynamicProgrammingAlgorithm(rootCounter->getCurrent(), true) ){
					subPatterns[x]->addItem(rootCounter->getCurrent());
					// prevCounter should never be null
					prevCounter->removeNextFromList();
					rootCounter = prevCounter->getNext();
				}else{
					prevCounter = rootCounter;
					rootCounter = rootCounter->getNext();
				}
			}

			addedYet = true;

		}else if(newIsSmaller && !newIsBigger){
			addPatternHelper(x,subPatterns[rootCounter->getCurrent()->getIndex()]);
			addedYet = true;
		}else if(newIsSmaller && newIsBigger){
			// equivalent!
			if(equivPatterns[rootCounter->getCurrent()->getIndex()] == NULL){
				equivPatterns[rootCounter->getCurrent()->getIndex()] = new patternlist(xmlArray[x]);
			}else{
				equivPatterns[rootCounter->getCurrent()->getIndex()]->addItem(xmlArray[x]);
			}
			shouldPrint[x] = rootCounter->getCurrent()->getIndex();
			addedYet = true;
		}else{
			// still have to call single-check function
			singleCheck(x, subPatterns[rootCounter->getCurrent()->getIndex()]);
		}
		prevCounter = rootCounter;
		if(rootCounter != NULL){
			rootCounter = rootCounter->getNext();
		}
	}
	if(!addedYet){
		if(prevCounter != NULL){
			prevCounter->addItem(xmlArray[x]);
		}else{
			rootPassedIn = new patternlist(xmlArray[x]);
		}
	}
};

void hierarchy::singleCheck(int x, patternlist* rootCounter)
{
//	patternlist* rootCounter = rootPassedIn;

	while (rootCounter != NULL){
		if(xmlArray[x]->DynamicProgrammingAlgorithm(rootCounter->getCurrent(), true) ){
			if(subPatterns[x] == NULL){
				subPatterns[x] = new patternlist(rootCounter->getCurrent());
			}else{
				subPatterns[x]->addItem(rootCounter->getCurrent());	
			}
		}else{
			singleCheck(x, subPatterns[rootCounter->getCurrent()->getIndex()]);
		}
		rootCounter = rootCounter->getNext();
	}
};


void hierarchy::printToDot(string filename, bool xPathsAsNumbers, bool showSource)
{

	ofstream dotFile;
	dotFile.open(filename.c_str());

	for (int i = 0; i < xmlSize; i++){
		
		dotFile << "//:: " << i << " ::: " << xmlArray[i]->getXPath() << endl;

	}

	dotFile << "digraph XPath_Query_Relationships {" << endl << endl;

	for (i = 0; i < xmlSize; i++){
		if(shouldPrint[i] < 0){
			
			int fieldLabel = 0;
			dotFile << i << " [shape=record," <<"label=\"{{<f" << fieldLabel << "1>XPath";
			if(xPathsAsNumbers){
				dotFile << " ID: " << i;
			}else{
				dotFile << ": " << xmlArray[i]->getPrintableXPath();
			}

			if(xmlArray[i]->getQueryNumber() >= 0){
				dotFile << "|<f" << fieldLabel << "2>XQuery ID: " << xmlArray[i]->getQueryNumber();
				if(showSource){
					dotFile << "|<f" << fieldLabel << "3>XQuery Part: " << xmlArray[i]->getXPathType();
				}
				dotFile << "}";
			}else{
				dotFile << "}";
			}

			patternlist* equivs = equivPatterns[i];

			while (equivs != NULL) {
				fieldLabel++;
				dotFile << "|{<f" << fieldLabel << "1>XPath";

				if(xPathsAsNumbers){
					dotFile << " ID: " << i;
				}else{
					dotFile << ": " << xmlArray[i]->getPrintableXPath();
				}

				if(xmlArray[i]->getQueryNumber() >= 0){
					dotFile << "|<f" << fieldLabel << "2>XQuery ID: " << xmlArray[equivs->getCurrent()->getIndex()]->getQueryNumber();
					if(showSource){
						dotFile << "|<f" << fieldLabel << "3>XQuery Part: " << xmlArray[equivs->getCurrent()->getIndex()]->getXPathType();
					}
					dotFile << "}";
				}else{
					dotFile << "}";
				}
				equivs = equivs->getNext();
			}

			dotFile << "}\"];" << endl;
		}
	}

	dotFile << endl;

	patternlist * rootItr = treeRoots;
//	while(rootItr !=NULL){}

	for (int j = 0; j < xmlSize; j++){
		
		rootItr = subPatterns[j];
		int first;
		int last;
		while(rootItr != NULL){
			if(shouldPrint[j] < 0){
				first = j;
			}else{
				first = shouldPrint[j];
			}
			if(shouldPrint[rootItr->getCurrent()->getIndex()] < 0){
				last = rootItr->getCurrent()->getIndex();
			}else{
				last = shouldPrint[rootItr->getCurrent()->getIndex()];
			}
			dotFile << first << " -> " << last << ";" << endl;
			rootItr = rootItr->getNext();
		}
	}

/*
			rootItr = equivPatterns[j];
			while(rootItr != NULL){
				if(shouldPrint[j] < 0){
					first = j;
				}else{
					first = shouldPrint[j];
				}
				if(shouldPrint[rootItr->getCurrent()->getIndex() < 0){
					last = rootItr->getCurrent()->getIndex()
				}else{
					last = shouldPrint[rootItr->getCurrent()->getIndex()];
				}
				dotFile << first << " -> " << last << ";" << endl;
				dotFile << last << " -> " << first << ";" << endl;
				rootItr = rootItr->getNext();
			}

*/

	dotFile << endl << "}";
	dotFile.close();


};

